# GNU Solfege - free ear training software
# vim: set fileencoding=utf-8 :
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import absolute_import

import hashlib
import logging
import os
import pickle
import sqlite3
import time
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from solfege import filesystem
from solfege import lessonfile
from solfege import mpd
from solfege import utils

import solfege


def hash_lessonfile_text(s):
    """
    Return the hashvalue of the string s, after filtering out:
    * lines starting with '#'
    * empty lines
    """
    lines = s.split("\n")
    lines = [x for x in lines if (not x.startswith("#") and bool(x))]
    sha1 = hashlib.sha1()
    sha1.update("\n".join(lines))
    return sha1.hexdigest()


def hash_of_lessonfile(filename):
    return hash_lessonfile_text(open(filename, 'r').read())


class DB(object):
    type_int_dict = {int: 0, unicode: 1, float: 2}
    int_type_dict = {0: int, 1: unicode, 2: float}
    class VariableTypeError(Exception): pass
    class VariableUndefinedError(Exception): pass
    """
    This DB class is currently located in src.statistics.py, but it might
    move to a separate/another file later, since it will store more data
    than just statistics.
   
    At startup on a fresh install:
    * Scan all lesson files (only standard files, not regression) to
      get uuid, filename, title, hash etc into the lessonfiles table.

    On every startup:
    * After the GUI is ready, we scan all lesson files in the background
      to check if any files have changed.

    When we start to practise an exercise:
    * Check if the lesson file has changed, and if so, we delete the statistics.
    """
    def __init__(self, callback=None, filename=None):
        """
        callback is called to display progress when scanning lesson files.
        """
        if filename:
            statistics_filename = filename
        else:
            try:
                if testsuite_is_running:
                    statistics_filename = ":memory:"
            except NameError:
                statistics_filename = self.get_statistics_filename()
        self.must_recreate = not os.path.exists(statistics_filename)
        try:
            if testsuite_is_running:
                self.must_recreate = False
        except NameError:
            pass
        self.conn = sqlite3.connect(statistics_filename)
        self.setup_tables()
        if self.must_recreate:
            self.scan_lessonfiles(callback)
            self.read_old_data()
    @staticmethod
    def get_statistics_filename():
        return os.path.join(filesystem.app_data(), "statistics.sqlite")
    def setup_tables(self):
        """
        lessonfiles table.
        ==================
        uuid: the uuid of the lesson file, as written in the lessonfile header.
        hash: hash value of the file the last time some statistics for the
              file was entered into the database.
        test_result: The result of the last time a test was made.
        test_passed: If the test was passed the last time taken.
        title: The lesson file, from the lesson file header
        module: The exercise module, taken from the lessonfile header
        """
        self.conn.execute('''create table if not exists lessonfiles
            (uuid text primary key not null,
             hash text not null,
             test_result float default 0.0,
             test_passed int default 0
            )''')
        columns = [x[1] for x in
              self.conn.execute('pragma table_info(lessonfiles)').fetchall()]
        if 'title' not in columns:
            self.conn.execute('alter table lessonfiles add column title txt')
        if 'module' not in columns:
            self.conn.execute('alter table lessonfiles add column module txt')
        if 'filename' not in columns:
            self.conn.execute('alter table lessonfiles add column filename txt')
        self.conn.execute('''create table if not exists sessions
            (uuid text, timestamp int, answerkey text, guessed text, count int, unique (uuid, timestamp, answerkey, guessed) )''')
        self.conn.execute('''create table if not exists tests
            (uuid text, timestamp int, answerkey text, guessed text, count int, unique (uuid, timestamp, answerkey, guessed) )''')
        self.conn.execute('''create table if not exists variables
            (variable_name text primary key not null,
             type int not null,
             value text not null)''')
    def get_field(self, uuid, fields):
        """
        Raise KeyError(uuid) if no lesson file with the uuid is in the
        database.
        """
        if not isinstance(fields, (tuple, list)):
            vfields = [fields]
        else:
            vfields = fields
        row = self.conn.execute('select %s from lessonfiles where uuid=?'
                                % ', '.join(vfields), (uuid,)).fetchone()
        if not row:
            raise KeyError(uuid)
        if isinstance(fields, (tuple, list)):
            return row
        else:
            return row[0]
    def scan_lessonfiles(self, callback=None):
        """
        Read info about standard lesson files into the lessonfiles table.
        Current lessonfile manager code cause us to parse the user contributed
        files in ~/lessonfiles too.
        """
        count = 0
        cc = lessonfile.manager.old_count()
        old_uuid = None
        uuid_count = 0
        progress = 0
        old_progress = 0
        for uuid, data in lessonfile.manager.parse(False):
            if callback and uuid != old_uuid:
                uuid_count += 1
                old_uuid = uuid
            self.set_lessonfile_data(uuid, data['header']['title'],
                    data['header']['module'], data['filename'])
            if callback:
                progress = uuid_count * 100 / cc
                if old_progress != progress:
                    callback(progress)
                    old_progress = progress
            count += 1
        logging.info("scan_lessonfiles: scanned %i files." % count)
    def set_lessonfile_data(self, uuid, title, module, filename):
        """
        Set the values title, module, filename and hash of uuid
        if the lesson file has changed.
        """
        row = self.conn.execute("select uuid, hash from lessonfiles where uuid=?",
                (uuid,)).fetchone()
        hashvalue = hash_of_lessonfile(filename)
        if row:
            if hashvalue != row[1]:
                self.validate_stored_statistics(uuid, filename)
        else:
            self.conn.execute("insert into lessonfiles "
                    "(uuid, title, module, filename, hash) "
                    "values (?, ?, ?, ?, ?)",
                    (uuid, title, module, filename, hashvalue))
        self.conn.commit()
    def read_old_data(self):
        """
        This function should be run once to import the old format statistics
        from ~/.solfege/statistics and into the sqlite database.
        We will import statistics for all uuids that exist in the lessonfiles table.
        """
        # Snippet from http://wiki.python.org/moin/UsingPickle/RenamingModules
        # that let us rename even though we have a renamed class
        renametable = {
            'src.dataparser': 'solfege.dataparser',
            }

        def mapname(name):
            if name in renametable:
                return renametable[name]
            return name

        def mapped_load_global(self):
            module = mapname(self.readline()[:-1])
            name = mapname(self.readline()[:-1])
            klass = self.find_class(module, name)
            self.append(klass)

        def load(file):
            unpickler = pickle.Unpickler(file)
            unpickler.dispatch[pickle.GLOBAL] = mapped_load_global
            return unpickler.load() 

        st_home = os.path.join(filesystem.app_data(), 'statistics')
        ign = 0
        imp = 0
        for row in self.conn.execute('select uuid from lessonfiles'):
            uuid = row[0]
            if not os.path.exists(os.path.join(st_home, uuid)):
                ign += 1
                continue
            for timestamp in os.listdir(os.path.join(st_home, uuid)):
                f = open(os.path.join(st_home, uuid, timestamp), 'r')
                session = load(f)
                f.close()
                for correct in session.keys():
                    for guess in session[correct]:
                        self.conn.execute("insert into sessions values(?, ?, ?, ?, ?)",
                        (uuid, timestamp, unicode(correct), unicode(guess), session[correct][guess]))
                imp += 1
        logging.info("imported statistics for %i lessonfiles, ignored %i" % (imp, ign))
        self.conn.commit()
    def last_test_result(self, uuid):
        """
        Return a float in the range 0.0 to 1.0 telling the score of the
        last test run on this uuid.
        Return None if no tests found.
        """
        cursor = self.conn.cursor()
        row = cursor.execute("select timestamp from tests where uuid=? order by -timestamp", (uuid,)).fetchone()
        # timestamp is not the timestamp of the last test run on this uuid.
        if row:
            timestamp = row[0]
            correct_count = cursor.execute("select sum(count) from tests where uuid=? and timestamp=? and answerkey=guessed", (uuid, timestamp)).fetchone()[0]
            total_count = cursor.execute("select sum(count) from tests where uuid=? and timestamp=?", (uuid, timestamp)).fetchone()[0]
            return correct_count * 1.0 / total_count
        return None
    def is_test_passed(self, uuid):
        row = self.conn.execute("select test_passed from lessonfiles "
                                "where uuid=?", (uuid,)).fetchone()
        if row:
            return row[0]
        return False
    def validate_stored_statistics(self, uuid, filename):
        """
        Check if the lesson file has changed since last time statistics for
        this lessonfile was added. Remove old statistics if the 
        m_t.m_P has already changed to the new file when this method is called.
        """
        cursor = self.conn.cursor()
        row = cursor.execute("select hash from lessonfiles where uuid=?", (uuid,)).fetchone()
        cur_lessonfile_hash_value = hash_of_lessonfile(filename)
        if not row:
            # Ususally the uuid exists in the database, but when running the
            # test suite, it does not, so we have the code here to add it.
            cursor.execute("insert into lessonfiles (uuid, hash) values(?, ?)",
                    (uuid, cur_lessonfile_hash_value))
            self.conn.commit()
        else:
            # row[0] is the hash value of the lesson file when all the existing
            # statistics for the lesson file with this uuid was saved.
            if row[0] != cur_lessonfile_hash_value:
                cursor.execute("delete from sessions where uuid=?", (uuid,))
                cursor.execute("update lessonfiles "
                    "set hash=?, test_passed=0, test_result=0.0 where uuid=?",
                    (cur_lessonfile_hash_value, uuid))
                self.conn.commit()
    def search_C(self, searchfor, columns):
        """
        Search for searchfor in the title column (and possible other places
        later). Return the columns named by the list columns.
        """
        searchfor = "%%%s%%" % searchfor
        n =  self.conn.execute("select %s " % ", ".join(columns)
             + "from lessonfiles where title like ?", (searchfor,)).fetchall()
        if not n:
            return []
        return [x for x in n if x[0] in solfege.lessonfile.manager.m_uiddb]
    def search(self, searchfor, columns):
        """
        Search for searchfor in the title column (and possible other places
        later). Return the columns named by the list columns.
        """
        searchfor = searchfor.lower()
        n =  self.conn.execute("select %s, title " % ", ".join(columns)
            + "from lessonfiles").fetchall()
        if not n:
            return []
        # The last conditional checks that the uuid exists in the lesson file
        # manager, since we don't want to return files that are not available
        # even thought there exist statistics for them.
        return [x[:-1] for x in n if searchfor in _(x[-1]) and x[0] in solfege.lessonfile.manager.m_uiddb]
    def search_tests(self, searchfor, columns):
        ret = self.search(searchfor, columns)
        ret = [r for r in ret if solfege.lessonfile.manager.get(r[0], 'test')]
        return ret
    def search_C_tests(self, searchfor, columns):
        ret = self.search_C(searchfor, columns)
        ret = [r for r in ret if solfege.lessonfile.manager.get(r[0], 'test')]
        return ret
    def recent(self, count):
        ids = []
        for row in self.conn.execute('select uuid, timestamp from sessions '
                'order by -timestamp'):
            if row[0] not in ids:
                ids.append(row[0])
            if len(ids) == count:
                break
        return ids
    def set_variable(self, name, value):
        """
        raise DB.VariableTypeError if the type we set is different from
        the one already stored.
        """
        row = self.conn.execute('select type, value from variables where variable_name=?', (name,)).fetchone()
        if row:
            saved_type = self.int_type_dict[row[0]]
        else:
            saved_type = type(value)
        if saved_type != type(value):
            raise DB.VariableTypeError()
        if type(value) not in self.type_int_dict:
            raise DB.VariableTypeError()
        if not row:
            self.conn.execute("insert into variables "
                    "(variable_name, type, value) "
                    "values (?, ?, ?)",
                    (name, self.type_int_dict[type(value)], unicode(value)))
        else:
            self.conn.execute("update variables "
                "set value=? where variable_name=?",
                (unicode(value), name))
    def get_variable(self, name):
        cursor = self.conn.execute('select type, value from variables where variable_name=?', (name,))
        try:
            type, value = cursor.fetchone()
        except TypeError:
            raise self.VariableUndefinedError()
        return self.int_type_dict[type](value)
    def del_variable(self, name):
        """
        Delete the variable, raise DB.VariableUndefinedError if the variable
        is not found.
        """
        cursor = self.conn.execute('delete from variables where variable_name=?', (name,))
        if not cursor.rowcount:
            # rowcount is 0 if no variables where deleted. We raise an 
            # error, since I'd like to know if we tries to do this.
            raise self.VariableUndefinedError()


class AbstractStatistics(object):
    def __init__(self, teacher):
        self.m_t = teacher
        self.m_timestamp = None
        self.m_test_mode = False
    def get_keys(self, all_keys=False):
        """
        Return the keys for all questions that have been answered correctly.
        If all_keys are false, it should also return the correct key for all
        questions that only have been answered wrongly.
        """
        if all_keys:
            c = solfege.db.conn.execute("select distinct(answerkey) from sessions where uuid=?", (self.m_t.m_P.header.lesson_id,))
        else:
            c = solfege.db.conn.execute("select distinct(answerkey) from sessions where uuid=? and answerkey=guessed", (self.m_t.m_P.header.lesson_id,))
        return [x[0] for x in c]
    def get_statistics(self, seconds):
        """
        return a dict with statistics more recent than 'seconds' seconds.
        The keys of dict are the correct answers for the lesson file.
        And the values of dict are new dicts where the keys are all the
        answers the user have given and the values are the number of times
        that particular answer has been given.
        """
        if seconds == -1:
            q = solfege.db.conn.execute("select answerkey, guessed, sum(count) from sessions where uuid=? group by answerkey, guessed", (self.m_t.m_P.header.lesson_id,))
        elif seconds == 0:
            q = solfege.db.conn.execute("select answerkey, guessed, sum(count) from sessions where uuid=? and timestamp=? group by answerkey, guessed", (self.m_t.m_P.header.lesson_id, self.m_timestamp))
        else:
            q = solfege.db.conn.execute("select answerkey, guessed, sum(count) from sessions where uuid=? and timestamp>? group by answerkey, guessed", (self.m_t.m_P.header.lesson_id, self.m_timestamp - seconds))
        ret = {}
        for answer, guess, count in q.fetchall():
            ret.setdefault(answer, {})
            ret[answer][guess] = count
        return ret
    def reset_session(self):
        """
        Start a new practise session.
        """
        self.m_timestamp = int(time.time())
    def enter_test_mode(self):
        self.m_test_mode = True
    def exit_test_mode(self):
        self.m_test_mode = False
        test_result = solfege.db.last_test_result(self.m_t.m_P.header.lesson_id)
        solfege.db.conn.execute("update lessonfiles "
                        "set test_result=?, test_passed=? where uuid=?",
            (test_result,
             test_result >= self.m_t.m_P.get_test_requirement(),
             self.m_t.m_P.header.lesson_id,))
        solfege.db.conn.commit()
    def _add(self, question, answer):
        """
        Register that for the question 'question' the user answered 'answer'.
        """
        assert self.m_timestamp
        # tuples must be converted to str to store them in sqlite.
        if isinstance(question, tuple):
            question = str(question)
        if isinstance(answer, tuple):
            answer = str(answer)
        table = {True: 'tests', False: 'sessions'}[self.m_test_mode]
        cursor = solfege.db.conn.cursor()
        row = cursor.execute(
                "select count from %s where uuid=? and timestamp=? "
                "and answerkey=? and guessed=?" % table,
                (self.m_t.m_P.header.lesson_id, 
                 self.m_timestamp, question, answer)).fetchone()
        if not row:
            cursor.execute(
                "insert into %s "
                "(uuid, timestamp, answerkey, guessed, count) "
                "values(?, ?, ?, ?, ?)" % table,
                (self.m_t.m_P.header.lesson_id, self.m_timestamp,
                 unicode(question), unicode(answer), 1))
        else:
            assert cursor.fetchone() is None
            cursor.execute(
                "update %s set count=? where "
                "uuid=? and timestamp=? and answerkey=? and guessed=?" % table,
                (row[0] + 1, self.m_t.m_P.header.lesson_id,
                 self.m_timestamp, question, answer))
        solfege.db.conn.commit()
    def add_wrong(self, question, answer):
        self._add(question, answer)
    def add_correct(self, answer):
        self._add(answer, answer)
    def get_last_test_result(self):
        """
        Return the test result of the last test ran for this lesson file.
        """
        return solfege.db.last_test_result(self.m_t.m_P.header.lesson_id)
    def get_percentage_correct(self):
        """Will return a 0 <= value <= 1.0 that say how many percent is
        correct in this session
        """
        num_correct = solfege.db.conn.execute("select sum(count) from sessions where answerkey=guessed and timestamp=? and uuid=?", (self.m_timestamp, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        num_asked = solfege.db.conn.execute("select sum(count) from sessions where timestamp=? and  uuid=?", (self.m_timestamp, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        if not num_correct:
            num_correct = 0
        if not num_asked:
            num_asked = 0
            return 0
        return 100.0 * num_correct / num_asked
    def get_percentage_correct_for_key(self, seconds, key):
        """
        Return the percentage correct answer the last 'seconds' seconds.
        """
        # All statistics
        num_guess = self.get_num_guess_for_key(seconds, key)
        if num_guess:
            return 100.0 * self.get_num_correct_for_key(seconds, key) / num_guess
        return 0
    def get_num_correct_for_key(self, seconds, key):
        """
        Return the number of correct answers for the given key 'key' the
        last 'seconds' seconds.
        Special meanings of 'seconds':
            -1  all statistics
             0  statistics from this session
        """
        if seconds == -1:
            ret = solfege.db.conn.execute("select sum(count) from sessions where answerkey=? and guessed=? and uuid=?", (key, key, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        elif seconds == 0:
            ret = solfege.db.conn.execute("select sum(count) from sessions where answerkey=? and guessed=? and timestamp=? and uuid=?", (key, key, self.m_timestamp, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        else:
            ret = solfege.db.conn.execute("select sum(count) from sessions where answerkey=? and guessed=? and timestamp>? and uuid=?", (key, key, self.m_timestamp - seconds, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        if ret:
            return ret
        return 0
    def get_num_guess_for_key(self, seconds, key):
        """
        See get_num_correct_for_key docstring.
        """
        if seconds == -1:
            ret = solfege.db.conn.execute("select sum(count) from sessions where answerkey=? and uuid=?", (key, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        elif seconds == 0:
            ret = solfege.db.conn.execute("select sum(count) from sessions where answerkey=? and timestamp=? and uuid=?", (key, self.m_timestamp, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        else:
            ret = solfege.db.conn.execute("select sum(count) from sessions where answerkey=? and timestamp>? and uuid=?", (key, self.m_timestamp - seconds, self.m_t.m_P.header.lesson_id)).fetchone()[0]
        if ret:
            return ret
        return 0


class LessonStatistics(AbstractStatistics):
    def key_to_pretty_name(self, key):
        def ff(x):
            try:
                t = eval(x)
                if type(t) == tuple:
                    return lessonfile.LabelObject(t[0], t[1])
            except Exception:
                return x
            return x
        for question in self.m_t.m_P.m_questions:
            if question.name.cval == key:
                return ff(question.name)
        return ff(key)

class IntervalStatistics(AbstractStatistics):
    def key_to_pretty_name(self, key):
        return utils.int_to_intervalname(int(key), 1, 1)

class HarmonicIntervalStatistics(AbstractStatistics):
    def key_to_pretty_name(self, key):
        return utils.int_to_intervalname(int(key), 1, 0)

class IdToneStatistics(LessonStatistics):
    def key_to_pretty_name(self, key):
        return mpd.MusicalPitch.new_from_notename(key).get_user_notename()





# GNU Solfege - free ear training software
# vim: set fileencoding=utf-8 :
# Copyright (C) 2006, 2007, 2008, 2009 Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Lesson file related GUI code.

from __future__ import absolute_import

############################
# Python Standard Library
############################
import gettext
import os
import warnings

###################
import gtk

###################
# Solfege modules
###################
from solfege import gu
from solfege import learningtree
from solfege import lessonfile
from solfege import mpd
import solfege

class CustomConfirmationAlert(gtk.Dialog):
    WIDTH_CHARS=70
    def __init__(self, buttons):
        gtk.Dialog.__init__(self, "", None, buttons=buttons)
        self.set_has_separator(False)
        self.set_resizable(False)
        self.set_border_width(6)
        self.vbox.set_spacing(12)
        hbox = gtk.HBox()
        hbox.set_spacing(12)
        hbox.set_border_width(6)
        self.vbox.pack_start(hbox, True)
        im = gtk.Image()
        im.set_alignment(0.5, 0.0)
        im.set_from_stock(gtk.STOCK_DIALOG_WARNING, gtk.ICON_SIZE_DIALOG)
        hbox.pack_start(im)
        # cvbox = contents vbox
        self.cvbox = gtk.VBox()
        hbox.pack_start(self.cvbox)
        self.g_primary = None
        self.g_secondary = None
    def set_markup(self, txt):
        self.m_primary = txt
        if not self.g_primary:
            self.g_primary = gtk.Label()
            self.g_primary.set_line_wrap(True)
            self.g_primary.set_width_chars(self.WIDTH_CHARS)
            self.g_primary.set_alignment(0.0, 0.5)
            self.cvbox.pack_start(self.g_primary, False)
        self.g_primary.set_markup("<b>%s</b>" % txt)
    def add_label(self, txt):
        label = gtk.Label(txt)
        label.set_width_chars(self.WIDTH_CHARS)
        label.set_line_wrap(True)
        label.set_alignment(0.0, 0.5)
        self.cvbox.pack_start(label, False)

class ConflictInfoDialog(CustomConfirmationAlert):
    def __init__(self, lesson_id, solfege_file, user_files):
        CustomConfirmationAlert.__init__(self,
                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                     gtk.STOCK_OK, gtk.RESPONSE_OK))
        self.set_default_response(gtk.RESPONSE_OK)
        # translators: 'lesson_id' is a source code variable name that should
        # not be translated.
        self.set_markup(_("The 'lesson_id' variable has to be unique for each lesson file. Solfege will now generate new 'lesson_id' for the conflicting files. One of the conflicting files must keep its 'lesson_id' because it belongs to GNU Solfege.") + " " + gettext.ngettext("Add new 'lesson_id' to the other file?", "Add new 'lesson_id' to the other files?", len(user_files)))
        self.add_label("")
        self.add_label(_("Press 'Cancel' to postpone the decision."))
        self.add_label("")
        self.add_label(u"Files to modify:")
        file_list = []
        for fn in user_files:
            self.add_label("    %s (%s)" % (fn['filename'], fn['timestr']))
        self.add_label("")
        self.add_label(_("Leave unmodified:"))
        self.add_label("    %s (%s)" % (
            os.path.abspath(solfege_file['filename']),
            solfege_file['timestr']))
        self.add_label("")
        self.show_all()


class ConflictResolveDialog(CustomConfirmationAlert):
    def __init__(self, lesson_id, selectable_files, info_files):
        CustomConfirmationAlert.__init__(self,
                buttons=(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                     gtk.STOCK_OK, gtk.RESPONSE_OK))
        self.set_markup(_("The 'lesson_id' variable has to be unique for each lesson file. Solfege will now generate new 'lesson_id' for the conflicting files. Please select which lesson file shall keep the current 'lesson_id'."))
        self.add_label("")
        self.add_label(_("The preselected file is Solfege's educated guess."))
        self.add_label(_("Press 'Cancel' to postpone the decision."))
        self.add_label("")
        first_btn = None
        oldest = sorted(selectable_files, key=lambda k: k['mtime'])[0]
        self.selected_filename = None
        for fileinfo in selectable_files:
            rdb = gtk.RadioButton(first_btn, "%s (%s)" % (
                os.path.abspath(fileinfo['filename']), fileinfo['timestr']),
                use_underline=False)
            def ff(w, f):
                self.selected_filename = f
            rdb.connect('clicked', ff, fileinfo['filename'])
            if first_btn is None:
                first_btn = rdb
            if fileinfo == oldest:
                rdb.set_active(True)
                self.selected_filename = fileinfo['filename']
            self.cvbox.pack_start(rdb, False)
        if info_files:
            self.add_label("")
            self.add_label(gettext.ngettext("This file is part of the same conflict, and will also get a new 'lesson_id':", "These files are part of the same conflict, and will also get a new 'lesson_id':", len(info_files)))
            for fileinfo in info_files:
                label = gtk.Label(u"    %(filename)s (%(timestr)s)" % fileinfo)
                label.set_alignment(0.0, 0.5)
                self.cvbox.pack_start(label, False)
        self.show_all()


def handle_lesson_id_crash(lessonfile_manager):
    """
    All lesson files that are part of Solfege, and distributed with the
    program are accessed by relative file name. In contrast are all the
    user contributed files in $HOME/lessonfiles accessed by absolute path.
    We call the solfege-file and user-file.

    A solfege-file will always keep its id if a user-file is in conflict.
    But with two solfege-files, or two user-files, Solfege will suggest
    to keep the oldest file, but let the user descide.
    """
    def is_write_protected(filename):
        return not (os.path.isfile(filename) and os.access(filename, os.W_OK))
    for lesson_id in lessonfile_manager.iterate_duplicated_lesson_id():
        infolist = list(lessonfile_manager.get_lesson_file_info(lesson_id))
        solfege_files = [d for d in infolist if not os.path.isabs(d['filename'])]
        user_files = [d for d in infolist if os.path.isabs(d['filename'])]
        del infolist
        if len(solfege_files) == 1:
            dlg = ConflictInfoDialog(lesson_id, solfege_files[0], user_files)
            ret = dlg.run()
            if ret == gtk.RESPONSE_OK:
                lessonfile_manager.duplicate_id_keep_this(lesson_id,
                        solfege_files[0]['filename'])
            elif ret in (gtk.RESPONSE_CANCEL, gtk.RESPONSE_DELETE_EVENT):
                    lessonfile_manager.ignore_duplicates_with_lesson_id(lesson_id)
            dlg.destroy()
        else:
            if len(solfege_files) > 1:
                dlg = ConflictResolveDialog(lesson_id, solfege_files, user_files)
            else:
                dlg = ConflictResolveDialog(lesson_id, user_files, [])
            ret = dlg.run()
            if ret == gtk.RESPONSE_OK:
                lessonfile_manager.duplicate_id_keep_this(lesson_id,
                            dlg.selected_filename)
            else:
                assert ret in (gtk.RESPONSE_CANCEL,
                               gtk.RESPONSE_DELETE_EVENT)
                lessonfile_manager.ignore_duplicates_with_lesson_id(lesson_id)
            dlg.destroy()


def rn_markup(s):
    v = []
    for p1, p2, p3, sep in lessonfile.rnc_markup_tokenizer(s):
        v.append('<span font_family="serif"><span size="xx-large">%s</span><span size="small">%s</span><span size="x-large">%s%s</span></span>' % (p1, p2, p3, sep))
    return "".join(v)

def chordname_markup(s):
    v = []
    for nn, mod, sup, bass in lessonfile.chordname_markup_tokenizer(s):
        if bass:
            bass = u"/%s" % mpd.MusicalPitch.new_from_notename(bass).get_user_notename()
        nn = mpd.MusicalPitch.new_from_notename(nn).get_user_notename()
        nn = nn[0].upper() + nn[1:]
        v.append('%s%s<span size="large" rise="11000">%s</span>%s' % (nn, mod, sup, bass))
    return u'<span font_family="serif" size="xx-large">%s</span>' % u" ".join(v)

def new_labelobject(label):
    """
    label can be one of the following types:
    * unicode string
    * lessonfile.LabelObject instance

    Return a gtk widget of some sort that displays the label.
    FIXME:
        If we in the future want some labels, for example chordlabel,
        to be transposable, then new_labelobject should return objects
        that are derived from gtk.Label, and they should have a
        a is_transposable method. Then Gui.on_new must call
        QuestionNameButtonTable and make it transpose the labels.
        But I don't think translated labels should have high priority.
    """
    if isinstance(label, basestring):
        # We hope all strings are unicode, but check for basestring just
        # in case some modules are wrong.
        l = gtk.Label(label)
        l.show()
    else:
        if isinstance(label, tuple):
            # I think only old code
            # that has  been removed. But let us keep the code until
            # we have a chanse to review.
            warnings.warn("lessonfilegui.new_labelobject: label is a tuple.",
                          DeprecationWarning)
            labeltype, labeldata = label
        else:
            labeltype = label.m_labeltype
            labeldata = label.m_labeldata
        if labeltype == 'progressionlabel':
            l = gu.HarmonicProgressionLabel(labeldata)
            l.show_all()
        else:
            l = gtk.Label()
            if labeltype == 'rnc':
                l.set_markup(rn_markup(labeldata))
            elif labeltype == 'chordname':
                l.set_markup(chordname_markup(labeldata))
            elif labeltype == 'plabel':
                l.set_markup("""<span font_family="serif"><span size="xx-large">%s</span><span size="small">%s</span><span size="x-large">%s</span></span>""" % labeldata)

            elif labeltype == 'pangomarkup':
                l.set_markup(labeldata)
            l.show()
    return l

class ExercisesMenuAddIn(object):
    def create_learning_tree_menu(self):
        """
        Create and return a gtk.Menu object that has submenus that
        let us select all lessons on the learning tree.
        """
        def create_menu(page):
            menu = gtk.Menu()
            for column in page:
                for section in column:
                    item = gtk.MenuItem(section.m_name)
                    for link in section:
                        if isinstance(link, learningtree.Page):
                            item = gtk.MenuItem(link.m_name)
                            menu.append(item)
                            item.set_submenu(create_menu(link))
                        else:
                            assert isinstance(link, unicode)
                            if solfege.db.get_field(link, 'module') not in ('melodicinterval', 'harmonicinterval', 'idbyname'):
                                continue
                            # We don't want to add these lesson files because we know
                            # that they cannot be exported. It would be better
                            # to catch these with a more generic algorithm, but
                            # then we would have to parse all the files, and that
                            # would be too slow.
                            if link in (
                                    # melodic-interval-self-config
                                    "f62929dc-7122-4173-aad1-4d4eef8779af",
                                    # harmonic-interval-self-config
                                    "466409e7-9086-4623-aff0-7c27f7dfd13b",
                                    # the csound-fifth-* files:
                                    "b465c807-d7bf-4e3a-a6da-54c78d5b59a1",
                                    "aa5c3b18-664b-4e3d-b42d-2f06582f4135",
                                    "5098fb96-c362-45b9-bbb3-703db149a079",
                                    "3b1f57e8-2983-4a74-96da-468aa5414e5e",
                                    "a06b5531-7422-4ea3-8711-ec57e2a4ce22",
                                    "e67c5bd2-a275-4d9a-96a8-52e43a1e8987",
                                    "1cadef8c-859e-4482-a6c4-31bd715b4787",
                                    ):
                                continue
                            item = gtk.MenuItem(solfege.db.get_field(link, 'title'))
                            item.connect('activate', self.on_select_exercise, link)
                            menu.append(item)
            return menu
                    
        menu = create_menu(solfege.app.m_frontpage_data)
        menu.show_all()
        self._menu_hide_stuff(menu)
        return menu
    def _menu_hide_stuff(self, menu):
        """
        Hide the menu if it has no menu items, or all menu items are hidden.
        """
        for sub in menu.get_children():
            assert isinstance(sub, gtk.MenuItem)
            if sub.get_submenu():
                self._menu_hide_stuff(sub.get_submenu())
                if not [c for c in sub.get_submenu().get_children() if c.get_property('visible')]:
                    sub.hide()




# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import errno
import gobject
import locale
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import webbrowser

from solfege import mpd
from solfege import soundcard

from solfege import abstract
from solfege import cfg
from solfege import dataparser
from solfege import gu
from solfege import i18n
from solfege import lessonfile
from solfege import osutils
from solfege import reportlib
from solfege import utils

import solfege

def check_rcfile():
    """See default.config for rcfileversion values, meanings and
    a description of how to add config variables.
    """
    rcfileversion = 15
    if cfg.get_int("app/rcfileversion") > rcfileversion:
        cfg.drop_user_config()
        return
    if cfg.get_int("app/rcfileversion") <= 1:
        if not "example-files" in cfg.get_string('config/lessoncollections'):
            cfg.set_string('config/lessoncollections', 
                "%s example-files" % cfg.get_string('config/lessoncollections'))
    if cfg.get_int("app/rcfileversion") <= 5:
        # This is more complicated that necessary to fix an old
        # error.
        if cfg.get_string("sound/commandline"):
            cfg.del_key("sound/commandline")
    if cfg.get_int("app/rcfileversion") <= 3:
        cfg.set_list("config/lessoncollections",
            cfg.get_string("config/lessoncollections").split())
    if cfg.get_int("app/rcfileversion") <= 4:
        cfg.del_key("config/web_browser")
    if sys.platform == 'win32':
        if cfg.get_string('sound/wav_player'):
            cfg.del_key('sound/wav_player')
    if cfg.get_int("app/rcfileversion") <= 5:
        cfg.set_string("mainwin/history_back_ak", "<alt>Left")
        cfg.set_string("mainwin/history_forward_ak", "<alt>Right")
        cfg.set_string("mainwin/history_reload_ak", "<ctrl>r")
    if cfg.get_int("app/rcfileversion") <= 6:
        cfg.set_list("config/lessoncollections", ['solfege', 'user'])
    if cfg.get_int("app/rcfileversion") <= 7:
        cfg.set_int("rhythm/countin_perc", 80)
    if cfg.get_int("app/rcfileversion") <= 8:
        cfg.del_key("singinterval/highest_tone")
        cfg.del_key("singinterval/lowest_tone")
        cfg.del_key("melodicinterval/highest_tone")
        cfg.del_key("melodicinterval/lowest_tone")
        cfg.del_key("harmonicinterval/highest_tone")
        cfg.del_key("harmonicinterval/lowest_tone")
    if cfg.get_int("app/rcfileversion") <= 9:
        cfg.del_section("mainwin")
    if cfg.get_int("app/rcfileversion") <= 10:
        cfg.del_section("lessoncollections")
        cfg.del_key("config/lessoncollections")
        for n in cfg.iterate_sections():
            cfg.del_key("%s/lessoncollection" % n)
            cfg.del_key("%s/lessonfile" % n)
    if cfg.get_int("app/rcfileversion") <= 11:
        for s in ('rhythm', 'rhythmtapping2'):
            cfg.del_key("%s/countin_perc" % s)
            cfg.del_key("%s/rhythm_perc" % s)
    if cfg.get_int("app/rcfileversion") <= 12:
        cfg.del_key("sound/card_info")
    if cfg.get_int("app/rcfileversion") <= 13:
        cfg.del_key("config/lowest_instrument_velocity")
        cfg.del_key("config/middle_instrument_velocity")
        cfg.del_key("config/highest_instrument_velocity")
        cfg.del_key("config/preferred_instrument_velocity")
    if cfg.get_int("app/rcfileversion") <= 14:
        # We have to split the midi_to_wav_cmd into two strings, and
        # moving the options to *_options, so that midi_to_wav_cmd only
        # have the name of the binary. This to allow binaries in dirs
        # with spaces.
        for k in ("midi_to_wav", "wav_to_mp3", "wav_to_ogg"):
            v = cfg.get_string("app/%s_cmd" % k).split(" ")
            cfg.set_string("app/%s_cmd" % k, v[0])
            cfg.set_string("app/%s_cmd_options" % k, " ".join(v[1:]))
    if cfg.get_int("app/rcfileversion") <= 15:
        for k in ("midi", "wav", "mp3", "ogg"):
            v = cfg.get_string("sound/%s_player" % k).split(" ")
            cfg.set_string("sound/%s_player" % k, v[0])
            cfg.set_string("sound/%s_player_options" % k,
                " ".join(v[1:]))
    cfg.set_int("app/rcfileversion", rcfileversion)
    try:
        a = mpd.notename_to_int(cfg.get_string("user/lowest_pitch"))
        b = mpd.notename_to_int(cfg.get_string("user/highest_pitch"))
    except mpd.InvalidNotenameException:
        if cfg.get_string("user/sex") == "male":
            cfg.set_string("user/highest_pitch", "e'")
            cfg.set_string("user/lowest_pitch", "c")
        else:
            cfg.set_string("user/highest_pitch", "e''")
            cfg.set_string("user/lowest_pitch", "c'")


class SolfegeApp(cfg.ConfigUtils):
    def __init__(self, options, lessonfile_manager):
        """
        options -- command line options parsed by optparse
        """
        cfg.ConfigUtils.__init__(self, 'solfege-app')
        self.m_startup_complete = False
        lessonfile.MusicBaseClass.temp_dir = tempfile.mkdtemp(prefix="solfege-")
        os.environ['SOLFEGETEMPDIR'] = lessonfile.MusicBaseClass.temp_dir
        # test_mode is when we are running a test from the Tests menu
        self.m_test_mode = False
        self.m_options = options
        self.m_teachers = {}
        self.m_running_exercise = None
        self.lessonfile_manager = lessonfile_manager
        self.m_sound_init_exception = None
        #
        self.m_userman_language = "C"
        for lang in i18n.langs():
            if os.path.isdir(os.path.join('help', lang)):
                self.m_userman_language = lang
                break
        # We have to check if lesson files have changed since last time we practised.
        self._pp = lessonfile_manager.parse(options.debug if options else None)
        gobject.idle_add(self.reparse_lessonfiles)
    def nonidle_reparse(self):
        total = len(self.lessonfile_manager.m_uiddb) * 2
        count = 0.0
        for uuid, data in self.lessonfile_manager.parse(self.m_options.debug):
            yield count / total
            count += 1
        from solfege import lessonfilegui
        lessonfilegui.handle_lesson_id_crash(lessonfile.manager)
        for uuid in self.lessonfile_manager.iterate_lesson_ids():
            data = lessonfile.manager.m_uiddb[uuid]
            solfege.db.set_lessonfile_data(uuid, data['header']['title'],
                    data['header']['module'], data['filename'])
            yield count / total
            count += 1
    def reparse_lessonfiles(self):
        """
        Parse all lesson files when we are idle.
        """
        # Check that solfege.win has been created. It is possible that
        # it is not created the first time we are idle, if the user is
        # prompted a question during MainWin construction
        if not getattr(solfege, 'win', None):
            return True
        try:
            uuid, data = self._pp.next()
            return True
        except StopIteration:
            solfege.win.handle_lesson_id_crash()
            self._pp = lessonfile.manager.iterate_lesson_ids()
            gobject.idle_add(self.check_for_changed_lessonfiles)
            return False
    def check_for_changed_lessonfiles(self):
        """
        Feed the parsed lesson file data to the database. We do
        it when we are idle.
        """
        try:
            uuid = self._pp.next()
            data = lessonfile.manager.m_uiddb[uuid]
            solfege.db.set_lessonfile_data(uuid, data['header']['title'],
                    data['header']['module'], data['filename'])
            return True
        except StopIteration:
            del self._pp
        self.m_startup_complete = True
        # If the user is really fast, or the computer not very idle, the
        # page displayed might need to be updated. It can contain warning
        # signs instead of links to some exercises.
        if solfege.win.m_viewer in ('frontpage', 'frontpage'):
            solfege.win.get_view().display_data(solfege.win.get_view()._page_tree)
        return False
    def setup_sound(self):
        if sys.platform == 'win32' and \
                    cfg.get_string("sound/type") == "sequencer-device":
            # just in case c:\home\.solfegerc is wrong
            cfg.set_string("sound/type", "winsynth")
        if self.m_options.no_sound \
           or cfg.get_string("sound/type") == "fake-synth":
            soundcard.initialise_using_fake_synth(self.m_options.verbose_sound_init)
        elif cfg.get_string("sound/type") == "winsynth":
            try:
                soundcard.initialise_winsynth(cfg.get_int("sound/synth_number"),
                      verbose_init=self.m_options.verbose_sound_init)
            except ImportError, e:
                self.display_sound_init_error_message(e)
                cfg.set_string("sound/type", "fake-synth")
                soundcard.initialise_using_fake_synth(True)
                return
            except RuntimeError, e:
                # We can get here if winmidi.output_devices() in winsynth
                # __init__ returns no devices. Don't know when, but it could
                # happen.
                solfege.win.display_exception_message(e)
                cfg.set_string("sound/type", "fake-synth")
                soundcard.initialise_using_fake_synth(True)
                return
            if cfg.get_int("sound/synth_number") != soundcard.synth.m_devnum:
                solfege.win.display_error_message2(_("MIDI setup"), _("MIDI Device %(olddev)i not available. Will use device %(newdev)i.") % {'olddev': cfg.get_int("sound/synth_number"), 'newdev': soundcard.synth.m_devnum})
                cfg.set_int("sound/synth_number", soundcard.synth.m_devnum)
        elif cfg.get_string("sound/type") == "external-midiplayer":
            soundcard.initialise_external_midiplayer(
                    cfg.get_string("sound/midi_player"),
                    verbose_init=self.m_options.verbose_sound_init)
            soundcard.synth.error_report_cb = solfege.win.display_error_message
        elif cfg.get_string("sound/type") == '':
            solfege.win.display_error_message(
                _("You should configure sound from the 'Sound' page "
                  "of the preferences window."))
        elif cfg.get_string("sound/type") == "sequencer-device":
            try:
                soundcard.initialise_devicefile(
                             cfg.get_string("sound/device_file"),
                             cfg.get_int("sound/synth_number"),
                             verbose_init=self.m_options.verbose_sound_init)
            except (soundcard.SoundInitException, OSError, ImportError), e:
                self.m_sound_init_exception = e
                soundcard.initialise_using_fake_synth(True)
        if cfg.get_string("programs/csound") == "AUTODETECT":
            for p in osutils.find_csound_executables():
                cfg.set_string("programs/csound", p)
                break
            else:
                # If not csound binary was found, then we set the string empty.
                # This means that autodetection will only happen the first time
                # you run the program. But later will newly installed binaries
                # be shown in the combo box of the preferences window.
                cfg.set_string("programs/csound", "")
        if cfg.get_string("programs/mma") == "AUTODETECT":
            for p in osutils.find_mma_executables(cfg.get_list("app/win32_ignore_drives")):
                cfg.set_string("programs/mma", p)
                break
            else:
                cfg.set_string("programs/mma", "")
    def display_sound_init_error_message(self, e):
        if isinstance(e, soundcard.SoundInitException):
            solfege.win.display_error_message(
            """%s""" % str(e).decode(locale.getpreferredencoding(), 'replace'))
        elif isinstance(e, ImportError):
            solfege.win.display_error_message2(str(e), _("You should configure sound from the preferences window, and try to use an external midi player. Or try to recompile the program and check for error messages to see why the module is not built."))
        elif e.errno == errno.EACCES:
            solfege.win.display_error_message(
                "The sound init failed: %s\n"
                "The errno EACCES indicates that you don't have write "
                "permission to the device."
                % str(e).decode(locale.getpreferredencoding(), 'replace'))
        elif e.errno == errno.EBUSY:
            solfege.win.display_error_message(
                "The sound init failed: %s\n"
                "It seems like some other program is using the device. You "
                "should try to quit that other program and restart Solfege."
                % str(e).decode(locale.getpreferredencoding(), 'replace'))
        else:
            solfege.win.display_error_message(
                "The sound init failed: %s\n"
                "You should configure sound from the 'Sound' page of "
                "the preferences window.\n\n"
                "It is also possible that the OS sound setup is incorrect."
                % str(e).decode(locale.getpreferredencoding(), 'replace'))
    def play_happy_sound(self):
        utils.play_music(r"\staff\relative c'{c16 e g a}", 180, 8,
               cfg.get_int('config/preferred_instrument_volume'))
    def play_sad_sound(self):
        utils.play_music(r"\staff\relative c'{<c,,8 cis>", 80, 58,
               cfg.get_int('config/preferred_instrument_volume'))
    def please_help_me(self):
        if isinstance(solfege.win.get_view(), abstract.Gui):
            # If the view visible is an exercise, when we check if the
            # lesson file header define a specific help page.
            if self.m_teachers[self.m_running_exercise].m_P.header.help:
                self.handle_href('%s.html' % self.m_teachers[self.m_running_exercise].m_P.header.help)
            else:
                # if not, we display help page named the same as the
                # exercise module
                self.handle_href('%s.html' % solfege.win.m_viewer)
    def show_exercise_theory(self):
        if self.m_teachers[self.m_running_exercise].m_P.header.theory:
            solfege.win.display_docfile("%s.html" % self.m_teachers[self.m_running_exercise].m_P.header.theory, "")
    def _practise_lesson_id(self, lesson_id, urlobj=None):
        """
        return the module name.
        """
        module = solfege.db.get_field(lesson_id, 'module')
        if self.m_running_exercise:
            solfege.win.box_dict[self.m_running_exercise].on_end_practise()
        if module not in self.m_teachers:
            self.create_teacher(module)
        if module not in solfege.win.box_dict:
            solfege.win.initialise_exercise(self.m_teachers[module])
        self.m_teachers[module].set_lessonfile(
            solfege.db.get_field(lesson_id, 'filename'))
        if self.m_teachers[module].m_P:
            solfege.win.activate_exercise(module, urlobj)
        self.m_running_exercise = module
        self.m_teachers[module].g_view = solfege.win.box_dict[module]
        solfege.win.show_help_on_current()
        return module
    def practise_lesson_id(self, lesson_id):
        def cleanup():
            module = solfege.db.get_field(lesson_id, 'module')
            self.m_teachers[module].m_P = None
            solfege.win.box_dict[module].practise_box.set_sensitive(False)
            solfege.win.box_dict[module].config_box.set_sensitive(False)
            solfege.win.box_dict[module].action_area.set_sensitive(False)
            solfege.win.box_dict[module].std_buttons_end_practise()
        try:
            module = self._practise_lesson_id(lesson_id)
        except (lessonfile.LessonfileParseException,
                dataparser.DataparserException), e:
            cleanup()
            solfege.win.display_exception_message(
                e, lessonfile=solfege.db.get_field(lesson_id, 'filename'))
            return
        # The try-except clause above will not catch trying to parse a
        # lesson_id that does not exist. This because
        # abstract.Teacher.parse_lessonfile will catch the exception and
        # display a message informing the user. So we have to check m_P here.
        # FIXME it can look as there is room for some redesign here...
        if not solfege.win.box_dict[module].m_t.m_P:
            cleanup()
            return
        if 'm_discards' in dir(self.m_teachers[module].m_P):
            for msg in self.m_teachers[module].m_P.m_discards:
                print >> sys.stderr, msg
        solfege.win.box_dict[module].practise_box.set_sensitive(True)
        solfege.win.box_dict[module].config_box.set_sensitive(True)
        solfege.win.box_dict[module].action_area.set_sensitive(True)
        solfege.win.box_dict[module].on_start_practise()
        w = solfege.win.g_ui_manager.get_widget("/Menubar/HelpMenu/PerExerciseHelp/HelpTheory")
        if w:
            w.set_sensitive(bool(self.m_teachers[module].m_P.header.theory))
        return module
    def test_lesson_id(self, lesson_id):
        self.m_test_mode = True
        module = self.practise_lesson_id(lesson_id)
        solfege.win.box_dict[module].on_start_practise()
        solfege.win.enter_test_mode()
    def handle_href(self, href, display_docfile_set_adj=1):
        # All generatet docbook documents are in utf-8
        urlobj = utils.Url(href)
        if not urlobj.protocol:
            if urlobj.filename.endswith(".midi"):
                if os.path.exists(os.path.join(solfege.win.box_dict['docviewer'].m_htmlwidget.m_document_wd, urlobj.filename)):
                    midifilename = urlobj.filename
                elif len(urlobj.filename.split('-')) > 1:
                    # Filename-nl.midi => Filename.midi
                    v = urlobj.filename.split('-')
                    w = v[-1].split('.')
                    midifilename = "-".join(v[:-1])+"."+w[1]
                else:
                    midifilename = urlobj.filename

                soundcard.play_mediafile("midi", os.path.join(
                    solfege.win.box_dict['docviewer'].m_htmlwidget.m_document_wd,
                    midifilename))
            else:
                solfege.win.display_docfile(urlobj.filename, urlobj.anchor)
        elif urlobj.protocol == 'solfege':
            self.handle_solfege_href(urlobj)
        elif urlobj.protocol == 'http':
            try:
                webbrowser.open_new(href)
            except Exception, e:
                solfege.win.display_error_message2(_("Error opening web browser"), str(e))
        elif urlobj.protocol == 'mailto':
            if self.get_string("config/mua"):
                try:
                    cmdline = self.get_string("config/mua") % urlobj.href
                except TypeError:
                    cmdline = "%s %s" % (self.get_string("config/mua"), urlobj.href)
                os.system(cmdline)
            else:
                gu.dialog_ok(_("You have not selected email program. You can do so in the preferences window."))
        else:
            gu.dialog_ok("unknown link type: %s" % urlobj.protocol)
    def handle_solfege_href(self, urlobj):
        """
        This method handles three kins of urls:
        solfege:practise/lesson_id
            practise the exercise with the lesson id 'lesson_id'
        solfege:practise/path/to/lessonfile
            Practise a specific lessonfile. Relative path:
                solfege:practise/lesson-files/chord-min-major
            Absolute path:
                solfege:practise//home/user/src/solfege/lesson-files/filename
        """
        solfege.win.hide_help_on_current()
        # the Url class does not know anything about lesson ids, but we
        # assume that lesson_ids don't include the '/' character.
        if '/' not in urlobj.lessonfile:
            self.practise_lesson_id(urlobj.lessonfile)
            return
        if urlobj.lessonfile:
            if not os.path.isfile(urlobj.lessonfile):
                solfege.win.display_error_message(_("Lessonfile not found:\n%s") % urlobj.lessonfile)
                return
        p = lessonfile.LessonfileCommon()
        p.parse_file(urlobj.lessonfile)
        self._practise_lesson_id(p.header.lesson_id, urlobj)
        # This loop changes variables in the lesson file header if they
        # have set values in the url.
        for key, val in urlobj.config.iteritems():
            try:
                self.m_teachers[p.header.module].m_P.header[key] = eval(val)
            except SyntaxError:
                gu.dialog_ok(_("Bad url parameter in link: '%s'.\nTrying to continue anyway.") % ("%s=%s" % (key, val)))
        solfege.win.box_dict[p.header.module].on_start_practise()
    def create_teacher(self, modulename):
        """
        Create the teacher in 'modulename' and add it to self.m_teachers.
        """
        m = __import__("solfege.exercises.%s" % modulename, fromlist=("solfege.exercises.%s" % modulename,), level=0)
        self.m_teachers[modulename] = m.Teacher(modulename)
    def reset_exercise(self, w=None):
        """
        Call on_end_practise, and then on_start_practise in
        the currently active exercise, if we have a exercise.
        """
        if isinstance(solfege.win.get_view(), abstract.Gui):
            solfege.win.get_view().on_end_practise()
            solfege.win.get_view().on_start_practise()
    def quit_program(self):
        try:
            cfg.sync()
        except IOError, e:
            solfege.win.display_exception_message(e)
        if isinstance(solfege.win.get_view(), abstract.Gui):
            g = solfege.win.get_view()
            # Check that the lesson file has a header, because if the
            # user tries to quit the program after parsing a bad lesson
            # file, we cannot call end_practise() without risking more
            # exceptions.
            if g.m_t.m_P and getattr(g.m_t.m_P, 'header', None):
                g.on_end_practise()
        solfege.db.conn.close()
        if soundcard.synth:
            soundcard.synth.close()
        shutil.rmtree(lessonfile.MusicBaseClass.temp_dir, True)
    def export_training_set(self, export_data, export_dir, output_format):
        """
        This function requires a program that can create WAV files
        from MIDI files and MP3 files from WAV.
        """
        def delay(n, tempo):
            """
            tempo is a dict of two integers
            """
            track = mpd.Track()
            track.set_bpm(*tempo)#self.get_int('config/default_bpm'))
            track.note(mpd.Rat(n, 4), 80, 0)
            soundcard.synth.play_track(track)
        track_idx = 0
        num = sum([x['count'] for x in export_data])
        # MainWin will set this to True if the user want to cancel
        # the export.
        self.m_abort_export = False
        report = reportlib.Report()
        report.append(reportlib.Heading(1, "Exported exercises"))
        table = reportlib.Table()
        report.append(table)
        for lesson_info in export_data:
            lesson_id = lesson_info['lesson_id']
            module = solfege.db.get_field(lesson_id, 'module')
            if module not in self.m_teachers:
                self.create_teacher(module)
            p = self.m_teachers[module].lessonfileclass()
            p.parse_file(solfege.db.get_field(lesson_id, 'filename'))
            for c in range(lesson_info['count']):
                if module == 'idbyname':
                    p.select_random_question()
                    if p.header.lesson_heading:
                        s = p.header.lesson_heading
                    else:
                        s = p.header.title
                    table.append_row("%i" % track_idx,
                                     p.get_question().name,
                                     s)
                    soundcard.start_export(os.path.join(
                            export_dir, "track-%i.mid" % track_idx))
                    for n in range(lesson_info.get('repeat', 1)):
                        p.play_question()
                        if n != lesson_info.get('repeat', 1) - 1:
                            if 'delay' in lesson_info:
                                delay(lesson_info['delay'], p.get_tempo())
                    soundcard.end_export()
                elif module in ('melodicinterval', 'harmonicinterval'):
                    t = self.m_teachers[module]
                    t.set_lessonfile(solfege.db.get_field(lesson_id, 'filename'))
                    t.start_practise()
                    t.new_question("c", "c''")
                    t.q_status = t.QSTATUS_SOLVED
                    try:
                        table.append_row("%i" % track_idx, "%s" % utils.int_to_intervalname(t.m_interval))
                    except AttributeError:
                        table.append_row("%i" % track_idx, "%s" % (" + ".join([utils.int_to_intervalname(q, False, True) for q in t.m_question])))
                    soundcard.start_export(os.path.join(
                            export_dir, "track-%i.mid" % track_idx))
                    for n in range(lesson_info.get('repeat', 1)):
                        t.play_question()
                        if n != lesson_info.get('repeat', 1) - 1:
                            if 'delay' in lesson_info:
                                delay(lesson_info['delay'],
                                    (self.get_int('config/default_bpm'), 4))
                    soundcard.end_export()
                else:
                    logging.warning("export_training_set:ignoring exercise with module='%s'" % module)
#####
                def do_convert(from_format, to_format):
                    """
                    Return False if we think the convert failed.
                    """
                    app_cfg_name = "app/%s_to_%s_cmd" % (from_format, to_format)
                    if from_format == 'midi':
                        from_ext = 'mid'
                    else:
                        from_ext = from_format
                    to_ext = to_format
                    if not cfg.get_string(app_cfg_name):
                        solfege.win.display_error_message2("Config variable not defined", "The missing or empty variable was '%s'" % app_cfg_name)
                        return False
                    try:
                        inout = {
                            'in': os.path.join(export_dir,
                                    "track-%i.%s" % (track_idx, from_ext)),
                            'out': os.path.join(export_dir,
                                    "track-%i.%s" % (track_idx, to_ext))}
                        opts = cfg.get_string(app_cfg_name + '_options').split(" ")
                        opts = [x % inout for x in opts]
                        # For some reasong setting the executable arg does
                        # not work for Python 2.5.2
                        try:
                            subprocess.call(
                                [cfg.get_string(app_cfg_name)] + opts)
                        except OSError, e:
                            raise osutils.BinaryForMediaConvertorException(app_cfg_name,
                                cfg.get_string(app_cfg_name), e)

                        if os.path.exists(os.path.join(export_dir, "track-%i.%s" % (track_idx, to_ext))):
                            os.remove(os.path.join(export_dir, "track-%i.%s" % (track_idx, from_ext)))
                        else:
                            # This means that the program failed to generate
                            # the WAV file. We set output_format to 'midi'
                            # because we don't want to display this error for
                            # every single file.
                            output_format = 'midi'
                            solfege.win.display_error_message2("External program must have failed", "The file in %(from)s format was not generated from the %(to)s file as expected. Please check your setup in the preferences window (CTRL-F12)." % {'to':to_format.upper(), 'from': from_format.upper()})
                    except (TypeError, KeyError):
                        solfege.win.display_error_message2("%(from)s to %(to)s config error", "There was a format string error. Will not generate WAV files. Please check the app/midi_to_wav_cmd config variable." % {'from': from_format, 'to': to_format})
                        output_format = 'midi'
                    return True
#####
                if output_format in ('mp3', 'wav', 'ogg'):
                    do_convert('midi', 'wav')
                if output_format in ('mp3', 'ogg'):
                    if not do_convert('wav', output_format):
                        output_format = 'wav'
                track_idx += 1
                yield 1.0 * track_idx / num
                if self.m_abort_export:
                    del self.m_abort_export
                    return
        reportlib.HtmlReport(report, os.path.join(export_dir, "toc.html"))
    def sheet_gen_questions(self, count, sdict):
        """
        count -- how many questions should we generate. We use this value
                 and not sdict['count'] because sometimes the app has some
                 questions, and just need a few more.
        """
        module = solfege.db.get_field(sdict['lesson_id'], 'module')
        if module not in self.m_teachers:
            self.create_teacher(module)
        p = self.m_teachers[module].lessonfileclass()
        p.parse_file(solfege.db.get_field(sdict['lesson_id'], 'filename'))
        if module == 'idbyname':
            for x in self._sheet_gen_question_idbyname(p, count, sdict):
                yield x
        else:
            assert module in ('harmonicinterval', 'melodicinterval')
            for x in self._sheet_gen_question_interval(module, p, count, sdict):
                yield x
    def _sheet_gen_question_idbyname(self, p, count, sdict):
        """
        yield count dicts, where each dict contain the data needed to
        print both the teachers and the students question.
        """
        counts = {}.fromkeys(range(len(p.m_questions)), 0)
        for x in range(count):
            while 1:
                p.select_random_question()
                if counts[p._idx] >= 1.0 * sdict['count'] / len(p.m_questions):
                    continue
                counts[p._idx] += 1
                break
            ret = {'question': {}, 'answer': {}}
            if sdict['qtype'] == 0:
                ret['question']['name'] = "...."
                ret['answer']['name'] = p.get_question().name
                ret['question']['music'] = p.get_lilypond_code()
                ret['answer']['music'] = p.get_lilypond_code()
                yield ret
            else:
                assert sdict['qtype'] == 1
                ret['question']['name'] = p.get_question().name
                ret['answer']['name'] = p.get_question().name
                ret['answer']['music'] = p.get_lilypond_code()
                ret['question']['music'] = p.get_lilypond_code_first_note()
                yield ret
    def _sheet_gen_question_interval(self, module, p, count, sdict):
        # FIXME in the idbyname we count how many times each question
        # has been selected, so get an even selection. We don't do it
        # here at the moment, because we need to descide what we really want.
        teacher = self.m_teachers[module]
        teacher.set_lessonfile(solfege.db.get_field(sdict['lesson_id'],
                                                          'filename'))
        teacher.start_practise()
        for x in range(count):
            teacher.new_question("c'", "c''")
            # quick hack to use this for both melodic and harmonic intervals
            if module == 'melodicinterval':
                teacher.m_interval = teacher.m_question[0]
            teacher.q_status = teacher.QSTATUS_SOLVED
            ret = {'question': {}, 'answer': {}}
            if sdict['qtype'] == 0:
                ret['question']['name'] = "...."
                ret['answer']['name'] = mpd.Interval.new_from_int(abs(teacher.m_interval)).get_name()
                ret['question']['music'] = r"\score{" \
                        r" { %s %s }" \
                        r"\layout { "\
                        r"  ragged-last = ##t " \
                        r"  \context { \Staff " \
                        r'\remove "Time_signature_engraver" } }' \
                        r"}" % (
                    teacher.m_tonika.get_octave_notename(),
                    (teacher.m_tonika + mpd.Interval.new_from_int(teacher.m_interval)).get_octave_notename())
                ret['answer']['music'] = ret['question']['music']
                yield ret
            else:
                assert sdict['qtype'] == 1
                ret['question']['name'] = mpd.Interval.new_from_int(abs(teacher.m_interval)).get_name()
                ret['answer']['name'] = mpd.Interval.new_from_int(abs(teacher.m_interval)).get_name()
                ret['question']['music'] = r"\score{ { %s s4 s4} "\
                       r"\layout{ "\
                       r"  ragged-last = ##t "\
                       r"  \context { \Staff "\
                       r'     \remove "Time_signature_engraver" } }'\
                       r"}" % teacher.m_tonika.get_octave_notename()
                ret['answer']['music'] = r"\score{ { %s %s } "\
                       r"\layout{ "\
                       r"  ragged-last = ##t "\
                       r"  \context { \Staff "\
                       r'     \remove "Time_signature_engraver" } }'\
                       r"}" % (
                    teacher.m_tonika.get_octave_notename(),
                    (teacher.m_tonika + teacher.m_interval).get_octave_notename())
                yield ret
    def load_learning_tree_as_frontpage_data(self, filename):
        """
        This function can read the old "learning tree" format datafile and
        convert it to a front page structure.
        Load into the .m_frontpage_data variable
        """
        data = {}
        data['collections'] = {}
        data['columns'] = [[]]

        d = eval(open(filename, "r").read())
        for k in d['menu']:
            data['columns'][0].append({'heading': k['name'].replace("_", ""), 'collection-ids': []})
            for k2 in k['children']:
                newkey = "%s-%s" % (k['name'].lower().replace(" ", "_"), k2['name'].lower().replace(" ", "_"))
                assert newkey not in data
                data['collections'][newkey] = k2
                data['columns'][0][-1]['collection-ids'].append(newkey)
                del k2['nametotranslate']
                k2['name'] = k2['name'].replace("_", "")

        lines = sum(len(x['collection-ids'])+1 for x in data['columns'][0])
        count = 0
        for idx, x in enumerate(data['columns'][0]):
            count += len(x['collection-ids']) + 1
            if count > lines / 2:
                if len(data['columns']) == 1:
                    data['columns'].append([])
                    for i in range(idx+1, len(data['columns'][0])):
                        data['columns'][1].append(data['columns'][0][idx+1])
                        del data['columns'][0][idx+1]
                    break
        self.m_frontpage_data = data



# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import absolute_import

import os
import codecs
import logging
import pprint
import sys

from solfege import filesystem
from solfege.gpath import Path

import solfege

def get_learning_tree_list(debug):
    """
    Return a list of tuples with info about the learning trees found.
    Tuple objects:
    1. The string "user" or "solfege"
    2. The filename, not including the name of the directory
    3. Full path to the learning tree.
    """
    trees = []
    for cls, dir in (('solfege', u'learningtrees'),
        ('user', os.path.join(filesystem.user_data(), "learningtrees"))):
        try:
            v = os.listdir(dir)
        except OSError:
            v = []
        for fn in v:
            if cls == 'solfege':
                fullpath = os.path.join("learningtrees", fn)
            else:
                fullpath = os.path.join(filesystem.user_data(), "learningtrees", fn)
            trees.append((cls, fn, fullpath))
    if not debug:
        try:
            trees.remove(('solfege', 'debugtree.txt', os.path.join('learningtrees', 'debugtree.txt')))
        except ValueError:
            # The debugtree.txtfile is for some reason missing
            pass
    return trees

class _TreeCommon(list):
    def __init__(self, items=[]):
        if isinstance(items, _TreeCommon):
            list.__init__(self, [items])
        else:
            list.__init__(self, items)
    def dump(self, stream, level=0):
        print >> stream, "%s%s([" % (" " * level, self.__class__.__name__)
        self.dump_children(stream, level)
    def dump_children(self, stream, level):
        for child in self:
            if isinstance(child, _TreeCommon):
                child.dump(stream, level + 1)
            else:
                print >> stream, "%su'%s'," % (" " * (level + 1), child)
        if level == 0:
            print >> stream, "%s])" % (" " * level)
        else:
            print >> stream, "%s])," % (" " * level)
    def iterate_lesson_ids(self):
        """
        Yield the lesson id of each lesson that has been added to this
        object or its children.
        """
        for child in self:
            if isinstance(child, _TreeCommon):
                for c in child.iterate_lesson_ids():
                    yield c
            if isinstance(child, (str, unicode)):
                yield child
    def get_use_dict(self):
        """
        Return a dict where the keys are lesson_ids of lessons used
        in this objects and its children. The values are an integer
        value telling how many times it have been used.
        """
        retval = {}
        for lesson_id in self.iterate_lesson_ids():
            retval[lesson_id] = retval.get(lesson_id, 0) + 1
        return retval
    def iterate_topics_for_id(self, lesson_id):
        for child in self:
            if isinstance(child, (Page, Column, LinkList)):
                for c in child.iterate_topics_for_id(lesson_id):
                    yield c
            elif child == lesson_id:
                yield self.m_name
    def iterate_flattened(self):
        for child in self:
            yield child
            if not isinstance(child, unicode):
                for subchild in child.iterate_flattened():
                    yield subchild
    @staticmethod
    def tests_in_sub(sub):
        if isinstance(sub, unicode):
            try:
                return bool(solfege.lessonfile.manager.get(sub, 'test'))
            except KeyError:
                return False
        for child in sub:
            if _TreeCommon.tests_in_sub(child):
                return True
        return False


class _NamedTreeCommon(_TreeCommon):
    def __init__(self, name=u'', listitems=[]):
        assert isinstance(name, unicode)
        _TreeCommon.__init__(self, listitems)
        self.m_name = name 
    def dump(self, stream, level=0):
        print >> stream, "%s%s(_(u'%s'), [" % (" " * level, self.__class__.__name__, self.m_name)
        self.dump_children(stream, level)
        
class LinkList(_NamedTreeCommon):
    """
    A list of links leading to exercises or new pages.
    """
    def append(self, item):
        assert isinstance(item, (str, unicode, Page))
        super(LinkList, self).append(item)
    def __str__(self):
        return "LinkList(_('%s')# len: %i)" % (self.m_name, len(self))

class Column(_TreeCommon):
    def add_linklist(self, heading):
        self.append(LinkList(heading))
        return self[-1]
    def __str__(self):
        return "Column(#len: %i)" % len(self)

class Page(_NamedTreeCommon):
    def __init__(self, name=u'', listitems=[]):
        assert isinstance(name, unicode)
        _NamedTreeCommon.__init__(self, name, listitems)
        self.m_modified = False
    def __repr__(self):
        return "<Page name=%s>" % self.m_name
    def get(self, path):
        """
        Return the element pointed to by path.
        """
        elem = self
        for idx in path[1:]:
            elem = elem[idx]
        return elem
    def is_empty(self):
        """
        Return True if all columns are empty or only contain empty sections.
        """
        for col in self:
            for sect in col:
                if isinstance(sect, LinkList) and len(sect) != 0:
                    return False
        return True

class Paragraph(object):
    def __init__(self, text):
        self.m_text = text


def convert_old_learningtree_data(data):
    """
    fn is the filename of a learningtree file. It reads the file as
    the default format was for solfege 3.14.
    Returns the Page objects
    """
    p = Page(_('Untitled%s' % ''), [Column()])
    col = p[-1]
    for d in data['menu']:
        linklist = col.add_linklist(unicode(d['name']))
        for section in d['children']:
            assert isinstance(section, dict)
            # We create a new page with 1 col for each set of links
            subpage = Page(unicode(section['name']))
            subpage.append(Column())
            subcol = subpage[-1]
            sublinklist = subcol.add_linklist(unicode(section['name']))
            for child in section['children']:
                sublinklist.append(unicode(child))
            linklist.append(subpage)
    return p

def parse_tree(s, C_locale=False):
    """
    This function can load both new (3.15.2) and older learning trees.
    This function will convert old format to new format, and always return
    new format files.
    """
    if C_locale:
        def ifu(s):
            # We do the isinstance check because this way we can handle
            # people manually editing the data file and adding strings that
            # are plain ascii, but not marked as unicode ( u'' )
            return s if isinstance(s, unicode) else unicode(s)
        ifunc = ifu
    else:
        ifunc = _
    namespace = {
        'Page': Page,
        'Column': Column,
        'LinkList': LinkList,
        'Paragraph': Paragraph,
        '_': ifunc,
        '__builtins__': {},
    }
    ret = eval(s, namespace, namespace)
    if isinstance(ret, dict):
        name = ret['title']
        ret = convert_old_learningtree_data(ret)
        ret.m_name = name
    return ret

def load_tree(fn, C_locale=False):
    return parse_tree(codecs.open(fn, "rU", 'utf-8', 'replace').read())




# GNU Solfege - free ear training software
# Copyright (C) 2009  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import absolute_import

import logging
import os
import StringIO
import sys
import time
import weakref

import gobject
import gtk

if __name__ == '__main__':
    from solfege import i18n
    i18n.setup(".", "C")
    import solfege.statistics
    solfege.db = solfege.statistics.DB()

import solfege
from solfege import cfg
from solfege import filesystem
from solfege import gu
from solfege import learningtree as pd
from solfege import lessonfilegui

class LessonFileDialogCommon(gtk.Dialog):
    def __init__(self, model):
        gtk.Dialog.__init__(self, _("Select lesson file"), self, buttons=
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,))
        self.m_model = model
        lessonfilegui.handle_lesson_id_crash(solfege.app.lessonfile_manager)
        self.g_ok_button = self.add_button("gtk-ok", gtk.RESPONSE_OK)
        self.g_ok_button.set_sensitive(False)
        self.set_default_size(800, 500)
        vbox = gtk.VBox()
        self.vbox.pack_start(vbox, False)
        self.g_liststore = gtk.ListStore(gobject.TYPE_STRING,
            gobject.TYPE_STRING, gobject.TYPE_STRING, gobject.TYPE_INT,
            gobject.TYPE_STRING)
        self.update_filelist()
        self.g_treeview = gtk.TreeView(self.g_liststore)
        self.g_treeview.connect('cursor-changed', self.on_cursor_changed)
        renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(_("Used"), renderer, text=3)
        column.set_sort_column_id(3)
        column.set_resizable(True)
        self.g_treeview.append_column(column)
        column = gtk.TreeViewColumn(_("Title"), renderer, text=0)
        column.set_resizable(True)
        column.set_sort_column_id(0)
        self.g_treeview.append_column(column)
        column = gtk.TreeViewColumn(_("Filename"), renderer, text=1)
        column.set_resizable(True)
        self.g_treeview.append_column(column)
        def sort_func(store, itera, iterb):
            return cmp(store.get(itera, 1)[0], store.get(iterb, 1)[0])
        self.g_treeview.get_model().set_sort_func(1, sort_func)
        column.set_sort_column_id(1)
        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.add(self.g_treeview)
        self.vbox.pack_start(scrolled_window)
        hbox = gtk.HBox()
        hbox.set_spacing(8)
        hbox.set_border_width(12)
        self.vbox.pack_start(hbox, False)
        hbox.pack_start(gtk.Label(_("Module:")), False)
        self.g_module = gtk.Label()
        hbox.pack_start(self.g_module, False)
        #
        hbox = gtk.HBox()
        hbox.set_spacing(8)
        hbox.set_border_width(12)
        self.vbox.pack_start(hbox, False)
        hbox.pack_start(gtk.Label(_("Used by topic(s):")), False)
        self.g_topics_vbox = gtk.VBox()
        hbox.pack_start(self.g_topics_vbox, False)
        #
        self.show_all()
        self.g_treeview.connect('row-activated', self.on_row_activated)
    def get_selected_lesson_id(self):
        """
        Returns None if no lesson file is selected.
        """
        i = self.g_treeview.get_cursor()[0]
        if i:
            i = i[0]
            iter = self.g_liststore.iter_nth_child(None, i)
            return self.g_liststore.get(iter, 2)[0]
    def on_row_activated(self, treeview, row, treeviewcolumn):
        self.response(gtk.RESPONSE_OK)
    def on_cursor_changed(self, treeview):
        i = self.g_treeview.get_cursor()[0][0]
        iter = self.g_liststore.iter_nth_child(None, i)
        self.g_module.set_text(self.g_liststore.get(iter, 4)[0])
        self.g_topics_vbox.foreach(lambda w: w.destroy())
        for s in self.m_model.iterate_topics_for_id(
                self.g_liststore.get(iter, 2)[0]):
            label = gtk.Label(s)
            label.set_alignment(0.0, 0.5)
            label.show()
            self.g_topics_vbox.pack_start(label)
        self.g_ok_button.set_sensitive(True)


class SelectLessonFileDialog(LessonFileDialogCommon):
    def __init__(self, model):
        LessonFileDialogCommon.__init__(self, model)
        hbox = gtk.HBox()
        hbox.set_spacing(gu.hig.SPACE_SMALL)
        hbox.show()
        self.vbox.pack_start(hbox, False)
        self.vbox.reorder_child(hbox, 0)
        g_show_used = gu.nCheckButton('SelectLessonFileDialog',
            'show_used_files', _("_Show files used in other topics"),
            False, self.update_filelist)
        hbox.pack_start(g_show_used, False)
        b = gtk.Button(_("_Search for new lesson files"))
        b.connect('clicked', self.search_for_lesson_files)
        b.show()
        hbox.pack_start(b, False)
        self.g_progress = gtk.ProgressBar()
        self.g_progress.set_text(_("Parsing files"))
        hbox.pack_start(self.g_progress, False)
    def search_for_lesson_files(self, btn):
        self.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
        self.g_progress.show()
        for percent in solfege.app.nonidle_reparse():
            self.g_progress.set_fraction(percent if percent <= 1.0 else 1.0)
            while gtk.events_pending():
                gtk.main_iteration()
        self.g_progress.hide()
        self.update_filelist()
        self.window.set_cursor(None)
    def update_filelist(self, w=None):
        self.g_liststore.clear()
        use_count = self.m_model.get_use_dict()
        if cfg.get_bool('SelectLessonFileDialog/show_used_files'):
            lesson_id_list = list(solfege.app.lessonfile_manager.iterate_lesson_ids())
        else:
            lesson_id_list = [i for i in list(solfege.app.lessonfile_manager.iterate_lesson_ids()) if use_count.get(i, 0) == 0]
        for lesson_id in lesson_id_list:
            d = solfege.app.lessonfile_manager.m_uiddb[lesson_id]
            self.g_liststore.append((d['header']['title'], d['filename'], lesson_id, use_count.get(lesson_id, 0),
                solfege.app.lessonfile_manager.get(lesson_id, 'module')))


def editor_of(obj):
    """
    Return the toplevel page, the one that is a Editor object.
    """
    p = obj
    while not isinstance(p, Editor):
        p = p.m_parent
    return p

def parent_page(obj):
    """
    Return the parent page of obj. Return None if this is the toplevel page.
    """
    p = obj
    while True:
        p = p.m_parent
        if isinstance(p, Page):
            return p
        if p is None:
            return None
    
class Section(gtk.VBox):
    """
    A section consists of a heading and a list of links.
    self.g_link_box is a vbox that contains the links.
    """
    def __init__(self, model, parent):
        gtk.VBox.__init__(self)
        self.m_model = model
        self.m_parent = parent
        assert isinstance(model, pd.LinkList)
        hbox = gtk.HBox()
        hbox.set_spacing(6)
        self.pack_start(hbox, False)
        # This is displayed and used when we edit the heading
        self.g_heading_entry = gtk.Entry()
        self.g_heading_entry.set_no_show_all(True)
        hbox.pack_start(self.g_heading_entry)
        self.g_heading = gtk.Label()
        self.g_heading.set_alignment(0.0, 0.5)
        # FIXME escape m_name
        self.g_heading.set_markup("<b>%s</b>" % model.m_name)
        hbox.pack_start(self.g_heading, False)
        #
        button_hbox = gtk.HBox()
        button_hbox.set_spacing(0)
        hbox.pack_start(button_hbox, False)
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_EDIT, gtk.ICON_SIZE_MENU)
        edit_btn = gtk.Button()
        edit_btn.add(im)
        edit_btn.connect('clicked', self.on_edit_heading)
        button_hbox.pack_start(edit_btn, False)
        #
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_ADD, gtk.ICON_SIZE_MENU)
        edit_btn = gtk.Button()
        edit_btn.add(im)
        edit_btn.connect('button-release-event', self.on_add)
        button_hbox.pack_start(edit_btn, False)
        #
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_CUT, gtk.ICON_SIZE_MENU)
        b = gtk.Button()
        b.add(im)
        b.connect('clicked', self.on_cut)
        button_hbox.pack_start(b, False)
        #
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_PASTE, gtk.ICON_SIZE_MENU)
        b = gtk.Button()
        b.add(im)
        b.connect('clicked', self.on_paste, -1)
        Editor.clipboard.register_paste_button(b, (pd.LinkList, pd.Page, unicode))
        button_hbox.pack_start(b, False)
        #
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_GO_DOWN, gtk.ICON_SIZE_MENU)
        self.g_move_down_btn = gtk.Button()
        self.g_move_down_btn.add(im)
        self.g_move_down_btn.connect('clicked', self.on_move_down)
        button_hbox.pack_start(self.g_move_down_btn, False)
        #
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_GO_UP, gtk.ICON_SIZE_MENU)
        self.g_move_up_btn = gtk.Button()
        self.g_move_up_btn.add(im)
        self.g_move_up_btn.connect('clicked', self.on_move_up)
        button_hbox.pack_start(self.g_move_up_btn, False)
        #
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_GO_BACK, gtk.ICON_SIZE_MENU)
        self.g_move_left_btn = gtk.Button()
        self.g_move_left_btn.add(im)
        self.g_move_left_btn.connect('clicked',
            parent.m_parent.on_move_section_left, self)
        button_hbox.pack_start(self.g_move_left_btn, False)
        #
        im = gtk.Image()
        im.set_from_stock(gtk.STOCK_GO_FORWARD, gtk.ICON_SIZE_MENU)
        self.g_move_right_btn = gtk.Button()
        self.g_move_right_btn.add(im)
        self.g_move_right_btn.connect('clicked',
            parent.m_parent.on_move_section_right, self)
        button_hbox.pack_start(self.g_move_right_btn, False)
        #
        self.g_link_box = gtk.VBox()
        self.pack_start(self.g_link_box, False)
        for link in self.m_model:
            self.g_link_box.pack_start(self.create_linkrow(link))
        # The button to click to add a new link
        hbox = gtk.HBox()
        self.pack_start(hbox)
    def on_move_down(self, btn):
        self.m_parent.move_section_down(self)
    def on_move_up(self, btn):
        self.m_parent.move_section_up(self)
    def on_edit_heading(self, btn):
        self.g_heading_entry.set_text(self.m_model.m_name)
        self.g_heading_entry.show()
        self.g_heading.hide()
        self.g_heading_entry.grab_focus()
        def finish_edit(entry):
            self.g_heading_entry.disconnect(sid)
            self.g_heading_entry.disconnect(keyup_id)
            self.g_heading_entry.disconnect(keydown_sid)
            self.m_model.m_name = entry.get_text()
            self.g_heading.set_markup(u"<b>%s</b>" % entry.get_text())
            self.g_heading_entry.hide()
            self.g_heading.show()
        sid = self.g_heading_entry.connect('activate', finish_edit)
        def keydown(entry, event):
            if event.keyval == gtk.keysyms.Tab:
                finish_edit(entry)
        keydown_sid = self.g_heading_entry.connect('key-press-event', keydown)
        def keyup(entry, event):
            if event.keyval == gtk.keysyms.Escape:
                self.g_heading_entry.disconnect(sid)
                self.g_heading_entry.disconnect(keyup_id)
                self.g_heading_entry.hide()
                self.g_heading.show()
                return True
        keyup_id = self.g_heading_entry.connect('key-release-event', keyup)
    def on_add(self, btn, event):
        menu = gtk.Menu()
        item = gtk.MenuItem(_("Add link to new page"))
        item.connect('activate', self.on_add_link_to_new_page)
        menu.append(item)
        item = gtk.MenuItem(_("Add link to exercise"))
        item.connect('activate', self.on_add_link)
        menu.append(item)
        menu.show_all()
        menu.popup(None, None, None, event.button, event.time)
    def on_add_link(self, btn):
        dlg = SelectLessonFileDialog(editor_of(self).m_model)
        ret = dlg.run()
        if ret in (gtk.RESPONSE_REJECT, gtk.RESPONSE_DELETE_EVENT, gtk.RESPONSE_CANCEL):
            pass
        else:
            assert ret == gtk.RESPONSE_OK
            self.m_model.append(dlg.get_selected_lesson_id())
            self.g_link_box.pack_start(
                self.create_linkrow(dlg.get_selected_lesson_id()), False)
        dlg.destroy()
    def on_add_link_to_new_page(self, menuitem):
        page = pd.Page(_("Untitled%s") % "", [pd.Column()])
        self.m_model.append(page)
        self.g_link_box.pack_start(self.create_linkrow(page))
    def create_linkrow(self, link_this):
        hbox = gtk.HBox()
        def ff(btn, page):
            if id(page) in editor_of(self).m_page_mapping:
                editor_of(self).show_page_id(id(page))
            else:
                if not page[0]:
                    page[0].append(pd.LinkList(link_this.m_name))
                p = Page(page, parent_page(self))
                p.show()
                editor_of(self).add_page(p)
        if isinstance(link_this, pd.Page):
            linkbutton = gu.ClickableLabel(link_this.m_name)
            linkbutton.connect('clicked', ff, link_this)
        else:
            linkbutton = gu.ClickableLabel(solfege.db.get_field(link_this, 'title'))
        hbox.pack_start(linkbutton)
        linkbutton.connect('button-press-event', self.on_right_click_row, link_this)
        hbox.show_all()
        return hbox
    def on_right_click_row(self, button, event, linked):
        idx = self.m_model.index(linked)
        if event.button == 3:
            m = gtk.Menu()
            item = gtk.ImageMenuItem(gtk.STOCK_DELETE)
            item.connect('activate', self.on_delete_link, linked)
            m.append(item)
            item = gtk.ImageMenuItem(gtk.STOCK_CUT)
            item.connect('activate', self.on_cut_link, idx)
            m.append(item)
            item = gtk.ImageMenuItem(gtk.STOCK_PASTE)
            item.set_sensitive(bool(Editor.clipboard))
            item.connect('activate', self.on_paste, idx)
            m.append(item)
            item = gtk.ImageMenuItem(gtk.STOCK_EDIT)
            item.connect('activate', self.on_edit_linktext, linked)
            item.set_sensitive(bool(not isinstance(linked, basestring)))
            m.append(item)
            m.show_all()
            m.popup(None, None, None, event.button, event.time)
            return True
    def on_delete_link(self, menuitem, linked):
        idx = self.m_model.index(linked)
        if id(linked) in editor_of(self).m_page_mapping:
            editor_of(self).destroy_window(id(linked))
        self.g_link_box.get_children()[idx].destroy()
        del self.m_model[idx]
    def on_edit_linktext(self, menuitem, linked):
        idx = self.m_model.index(linked)
        # row is the hbox containing the linkbutton
        row = self.g_link_box.get_children()[idx]
        linkbutton = row.get_children()[0]
        entry = gtk.Entry()
        entry.set_text(linkbutton.get_label())
        row.pack_start(entry)
        linkbutton.hide()
        entry.show()
        entry.grab_focus()
        def finish_edit(entry):
            linkbutton.set_label(entry.get_text().decode("utf-8"))
            linkbutton.get_children()[0].set_alignment(0.0, 0.5)
            linkbutton.show()
            self.m_model[idx].m_name = entry.get_text().decode("utf-8")
            entry.destroy()
        sid = entry.connect('activate', finish_edit)
        def keydown(entry, event):
            if event.keyval == gtk.keysyms.Tab:
                finish_edit(entry)
        entry.connect('key-press-event', keydown)
        def keyup(entry, event):
            if event.keyval == gtk.keysyms.Escape:
                linkbutton.show()
                entry.disconnect(sid)
                entry.destroy()
                return True
        entry.connect('key-release-event', keyup)
    def on_cut(self, btn):
        self.m_parent.cut_section(self)
    def on_cut_link(self, menuitem, idx):
        Editor.clipboard.append(self.m_model[idx])
        del self.m_model[idx]
        self.g_link_box.get_children()[idx].destroy()
    def on_paste(self, btn, idx):
        assert Editor.clipboard, "Paste buttons should be insensitive when the clipboard is empty."
        pobj = Editor.clipboard.pop()
        if isinstance(pobj, pd.LinkList):
            mobj = pd.Page(pobj.m_name, [pd.Column(pobj)])
        else:
            mobj = pobj
        if idx == -1:
            self.m_model.append(mobj)
            self.g_link_box.pack_start(self.create_linkrow(mobj))
        else:
            self.m_model.insert(idx, mobj)
            row = self.create_linkrow(mobj)
            self.g_link_box.pack_start(row)
            self.g_link_box.reorder_child(row, idx)


class Column(gtk.VBox):
    def __init__(self, model, parent):
        gtk.VBox.__init__(self)
        self.set_spacing(gu.hig.SPACE_MEDIUM)
        self.m_model = model
        self.m_parent = parent
        assert isinstance(model, pd.Column)
        self.g_section_box = gtk.VBox()
        self.g_section_box.set_spacing(gu.hig.SPACE_MEDIUM)
        self.pack_start(self.g_section_box, False)
        for section in model:
            assert isinstance(section, pd.LinkList)
            gui_section = Section(section, self)
            self.g_section_box.pack_start(gui_section, False)
        hbox = gtk.HBox()
        self.pack_start(hbox, False)
        b = gtk.Button(_("Add section"))
        hbox.pack_start(b, False)
        b.connect('clicked', self.on_add_section)
        b = gtk.Button(stock=gtk.STOCK_PASTE)
        b.connect('clicked', self.on_paste)
        Editor.clipboard.register_paste_button(b, pd.LinkList)
        hbox.pack_start(b, False)
    def __del__(self):
        logging.debug("Column.__del__")
    def cut_section(self, section):
        idx = self.g_section_box.get_children().index(section)
        Editor.clipboard.append(self.m_model[idx])
        del self.m_model[idx]
        self.g_section_box.get_children()[idx].destroy()
    def on_add_section(self, btn):
        # We write "Untitled%s" % "" instead of just "Untitled" here
        # since "Untitled%s" is already translated in many languages.
        section = pd.LinkList(_("Untitled%s" % ""))
        self.m_model.append(section)
        gui_section = Section(section, self)
        self.g_section_box.pack_start(gui_section, False)
        gui_section.show_all()
    def move_section_down(self, section):
        idx = self.g_section_box.get_children().index(section)
        if idx < len(self.g_section_box.get_children()) - 1:
            self.g_section_box.reorder_child(section, idx + 1)
            self.m_model[idx], self.m_model[idx + 1] \
                    = self.m_model[idx + 1], self.m_model[idx]
            self.m_parent.update_buttons()
    def move_section_up(self, section):
        idx = self.g_section_box.get_children().index(section)
        if idx > 0:
            self.g_section_box.reorder_child(section, idx - 1)
            self.m_model[idx], self.m_model[idx - 1] \
                    = self.m_model[idx - 1], self.m_model[idx]
            self.m_parent.update_buttons()
    def on_paste(self, widget):
        """
        Paste the clipboard as a new section to this column.
        """
        assert Editor.clipboard, "Paste buttons should be insensitive when the clipboard is empty."
        assert isinstance(Editor.clipboard[-1], pd.LinkList)
        pobj = Editor.clipboard.pop()
        self.m_model.append(pobj)
        sect = Section(pobj, self)
        sect.show_all()
        self.g_section_box.pack_start(sect, False)


class Page(gtk.VBox):
    def __init__(self, model, parent):
        gtk.VBox.__init__(self)
        self.m_model = model
        self.m_parent = parent
        sc = gtk.ScrolledWindow()
        sc.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.pack_start(sc)
        self.g_column_box = gtk.HBox()
        self.g_column_box.set_spacing(gu.hig.SPACE_LARGE)
        self.g_column_box.set_border_width(gu.hig.SPACE_SMALL)
        # We pack column into this box
        sc.add_with_viewport(self.g_column_box)
        self.show_all()
        if model:
            self.update_from_model()
    def __del__(self):
        logging.debug("Page.__del__:", self.m_model.m_name)
    def on_add_column(self, *btn):
        column = pd.Column()
        self.m_model.append(column)
        gcol = Column(column, self)
        gcol.show_all()
        self.g_column_box.pack_start(gcol)
    def on_move_section_left(self, button, section):
        column_idx = self.g_column_box.get_children().index(section.m_parent)
        section_idx = section.m_parent.g_section_box.get_children().index(section)
        if column_idx > 0:
            to_column = self.g_column_box.get_children()[column_idx - 1]
            section.reparent(to_column.g_section_box)
            section.m_parent = to_column
            to_column.g_section_box.set_child_packing(section, False, False, 0, gtk.PACK_START)
            self.m_model[column_idx - 1].append(self.m_model[column_idx][section_idx])
            del self.m_model[column_idx][section_idx]
            # Remove the right-most column if we moved the
            # last section out of it.
            if not self.g_column_box.get_children()[-1].g_section_box.get_children():
                assert len(self.m_model[-1]) == 0
                del self.m_model[-1]
                self.g_column_box.get_children()[-1].destroy()
            self.update_buttons()
    def on_move_section_right(self, button, section):
        # the column we move from
        column_idx = self.g_column_box.get_children().index(section.m_parent)
        section_idx = section.m_parent.g_section_box.get_children().index(section)
        if column_idx == len(self.g_column_box.get_children()) - 1:
            self.on_add_column()
        to_column = self.g_column_box.get_children()[column_idx + 1]
        section.reparent(to_column.g_section_box)
        section.m_parent = to_column
        to_column.g_section_box.set_child_packing(section, False, False, 0, gtk.PACK_START)
        to_section_idx = len(self.m_model[column_idx + 1])
        self.m_model[column_idx + 1].append(self.m_model[column_idx][section_idx])
        del self.m_model[column_idx][section_idx]
        self.update_buttons()
    def update_from_model(self):
        for child in self.g_column_box.get_children():
            child.destroy()
        for column in self.m_model:
            self.g_column_box.pack_start(Column(column, self))
        self.g_column_box.show_all()
        self.update_buttons()
    def update_buttons(self):
        num_cols = len(self.g_column_box.get_children())
        for col_idx, column in enumerate(self.g_column_box.get_children()):
            num_sects = len(column.g_section_box.get_children())
            for sect_idx, section in enumerate(column.g_section_box.get_children()):
                section.g_move_up_btn.set_sensitive(sect_idx != 0)
                section.g_move_down_btn.set_sensitive(sect_idx != num_sects -1)
                section.g_move_left_btn.set_sensitive(col_idx != 0)
                if [col for col in self.g_column_box.get_children() if not col.g_section_box.get_children()] and col_idx == num_cols - 1:
                    section.g_move_right_btn.set_sensitive(False)
                else:
                    section.g_move_right_btn.set_sensitive(True)


class Clipboard(list):
    def __init__(self, v=[]):
        list.__init__(v)
        self.m_paste_buttons = []
    def pop(self, i=-1):
        ret = list.pop(self, i)
        self.update_buttons()
        return ret
    def append(self, obj):
        list.append(self, obj)
        self.update_buttons()
    def register_paste_button(self, button, accepts_types):
        self.m_paste_buttons.append((button, accepts_types))
    def update_buttons(self):
        for button, types in self.m_paste_buttons:
            button.set_sensitive(bool(self) and isinstance(self[-1], types))


class Editor(gtk.Window, gu.EditorDialogBase):
    savedir = os.path.join(filesystem.user_data(), 'learningtrees')
    # The clipboard will be shared between all Editor instances
    clipboard = Clipboard()
    def __init__(self, filename=None):
        gtk.Window.__init__(self)
        logging.debug("fpeditor.Editor.__init__(%s)" % filename)
        gu.EditorDialogBase.__init__(self, filename)
        self.set_default_size(800, 600)
        self.g_main_box = gtk.VBox()
        self.add(self.g_main_box)
        self.g_actiongroup.add_actions([
            ('GoBack', gtk.STOCK_GO_BACK, None, None, None, self.go_back),
        ])
        self.setup_toolbar()
        # This dict maps the windows created for all pages belonging to
        # the file.
        self.m_page_mapping = {}
        self.m_model = None
        if filename:
            self.load_file(filename)
        else:
            self.m_model = pd.Page(_("Untitled%s") % self.m_instance_number,
                    pd.Column())
            self.set_not_modified()
        self.m_page_mapping[id(self.m_model)] = Page(self.m_model, self)
        self.g_visible_page = self.m_page_mapping[id(self.m_model)]
        self.g_main_box.pack_start(self.m_page_mapping[id(self.m_model)])
        self.clipboard.update_buttons()
        self.show_all()
    def __del__(self):
        logging.debug("fpeditor.Editor.__del__, filename=" % self.m_filename)
    def add_page(self, page):
        """
        Add and show the page.
        """
        editor_of(self).m_page_mapping[id(page.m_model)] = page
        self.g_main_box.pack_start(page)
        self.show_page(page)
    def show_page_id(self, page_id):
        self.show_page(self.m_page_mapping[page_id])
    def show_page(self, page):
        """
        Hide the currently visible page, and show PAGE instead.
        """
        self.g_visible_page.hide()
        self.g_visible_page = page
        page.show()
        self.g_ui_manager.get_widget("/Toolbar/GoBack").set_sensitive(
            not isinstance(self.g_visible_page.m_parent, Editor))
    def go_back(self, *action):
        self.show_page(self.g_visible_page.m_parent)
    def setup_toolbar(self):
        self.g_ui_manager.insert_action_group(self.g_actiongroup, 0)
        uixml = """
        <ui>
         <toolbar name='Toolbar'>
          <toolitem action='GoBack'/>
          <toolitem action='New'/>
          <toolitem action='Open'/>
          <toolitem action='Save'/>
          <toolitem action='SaveAs'/>
          <toolitem action='Close'/>
          <toolitem action='Help'/>
         </toolbar>
         <accelerator action='Close'/>
         <accelerator action='New'/>
         <accelerator action='Open'/>
         <accelerator action='Save'/>
        </ui>
        """
        self.g_ui_manager.add_ui_from_string(uixml)
        toolbar = self.g_ui_manager.get_widget("/Toolbar")
        self.g_main_box.pack_start(toolbar, False)
        self.g_main_box.reorder_child(toolbar, 0)
        self.g_ui_manager.get_widget("/Toolbar").set_style(gtk.TOOLBAR_BOTH)
    def destroy_window(self, window_id):
        """
        Destroy the window with the id 'windowid' and all subwindows.
        """
        def do_del(wid):
            for key in [k for k in self.m_page_mapping.keys()
                    if id(parent_page(self.m_page_mapping[k]).m_model) == wid]:
                do_del(key)
            editor_of(self).m_page_mapping[wid].destroy()
            del editor_of(self).m_page_mapping[wid]
        do_del(window_id)
    @staticmethod
    def edit_file(fn):
        if fn in Editor.instance_dict:
            Editor.instance_dict[fn].present()
        else:
            try:
                win = Editor(fn)
            except IOError, e:
                gu.dialog_ok(_("Loading file '%(filename)s' failed: %(msg)s") %
                        {'filename': fn, 'msg': str(e).decode('utf8', 'replace')})
            win.show()
    @staticmethod
    def tuple_to_filename(place, filename):
        if place == 'solfege':
            return os.path.join("learningtrees", filename)
        else:
            assert place == 'user'
            return os.path.join(filesystem.user_data(), 'learningtrees', filename)
    def load_file(self, filename):
        """
        Load a file into a empty, newly created Editor object.
        """
        assert self.m_model == None
        self.m_model = pd.load_tree(filename, C_locale=True)
        self.m_filename = filename
        # 
        if not os.path.isabs(filename):
            if not os.access(filename, os.W_OK):
                m = gtk.MessageDialog(self, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO,
                    gtk.BUTTONS_CLOSE, _("The default learning tree is write protected in your install. This is normal. If you want to edit a learning tree, you have to select one of the trees stored in .solfege/learningtrees in your home directory."))
                m.run()
                m.destroy()
        self.set_not_modified()
        self.set_title(self.m_filename)
    def set_not_modified(self):
        """
        Store the current state of the data in self.m_orig_dump so that
        is_modified() will return False until we make new changes.
        """
        io = StringIO.StringIO()
        self.m_model.dump(io)
        self.m_orig_dump = io.getvalue()
    def is_modified(self):
        """
        Return True if the data has changed since the last call to
        set_not_modified()
        """
        io = StringIO.StringIO()
        self.m_model.dump(io)
        s = io.getvalue()
        return s != self.m_orig_dump
    @property
    def m_changed(self):
        return self.is_modified()
    def save(self, w=None):
        assert self.m_filename
        f = open(self.m_filename, 'w')
        self.m_model.dump(f)
        f.close()
        self.set_not_modified()
        # We do test for solfege.win since it is not available during testing
        if getattr(solfege, 'win', None):
            solfege.win.on_learning_tree_changed()
    def on_show_help(self, *w):
        return

if __name__ == '__main__':
    gtk.link_button_set_uri_hook(lambda a, b: None)
    e = Editor()
    e.load_file("learningtrees/learningtree.txt")
    gtk.main()
